using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.ServiceModel;
using System.ServiceModel.Channels;
using Westwind.Tools;

namespace WebConnectionWcf
{
    [ComVisible(true)]
    [ProgId("WebConnectionWcf.WebConnectionWcfClient")]
    public class WebConnectionWcfClient
    {
        private string endPointUri = "net.pipe://localhost/WebConnectionDemo";

        /// <summary>
        /// Error message returned on a failure result from method calls
        /// </summary>
        public string ErrorMessage
        {
            get { return _ErrorMessage; }
            set { _ErrorMessage = value; }
        }
        private string _ErrorMessage = "";

        /// <summary>
        /// The internal Proxy instance used for this 
        /// </summary>
        protected IWebConnectionService Proxy
        {
            get
            {
                if (_Proxy == null)
                {
                    try
                    {
                        EndpointAddress ep = new EndpointAddress(this.endPointUri);

                        Binding Binding = null;
                        if (this.endPointUri.StartsWith("net.tcp://"))
                            Binding = new NetTcpBinding();
                        else if (this.endPointUri.StartsWith("net.pipe://"))
                            Binding = new NetNamedPipeBinding();
                        else if (this.endPointUri.StartsWith("http://"))
                            Binding = new BasicHttpBinding();

                        _Proxy = ChannelFactory<IWebConnectionService>.CreateChannel(Binding, ep);
                    }
                    catch (Exception ex)
                    {
                        this.ErrorMessage = ex.Message;
                    }
                }
                return _Proxy;
            }
            set { _Proxy = value;}
        }
        private IWebConnectionService _Proxy = null;


        /// <summary>
        /// Allow client to specify the endpoint - can't do
        /// constructor over COM so we need a method here
        /// </summary>
        /// <param name="EndPointUri"></param>
        /// <returns></returns>
        public bool SetEndPoint(string EndPointUri)
        {
            this.endPointUri = EndPointUri;
            return true;
        }

        /// <summary>
        /// Process a Web Connection Server request on the server
        /// </summary>
        /// <param name="RequestData"></param>
        /// <returns></returns>
        public string ProcessHit(string RequestData)
        {
            string Output = null;

            if (this.Proxy == null)
                return null;

            try
            {

                Output = Encoding.Default.GetString(Proxy.ProcessHit("Query_String=wwdemo~TestPage") as byte[]);
            }
            catch (Exception ex)
            {
                this.ErrorMessage = ex.Message;
            }

            return Output;
        }

        /// <summary>
        /// Gets the ProcessID from the Web Connection Server
        /// </summary>
        /// <param name="InstanceId"></param>
        /// <returns></returns>
        public int GetProcessId(int InstanceId)
        {
            return Proxy.GetProcessId(InstanceId);
        }

        /// <summary>
        /// Cleanup the Proxy if not in use
        /// </summary>
        public void ReleaseProxy()
        {
            this.Proxy = null;
        }
    }
}
