using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.ServiceModel;
using System.ServiceModel.Channels;
using Westwind.Tools;

namespace WebConnectionWcf
{
    [ComVisible(true)] 
    [ProgId("WebConnectionWcf.WebConnectionWcfHost")]
    public class WebConnectionWcfHost : IDisposable
    {
        ServiceHost Host = null;
        object WebConnectionServer = null;
        
        public string ErrorMessage
        {
            get { return _ErrorMessage; }
            set { _ErrorMessage = value; }
        }
        private string _ErrorMessage = "";

        /// <summary>
        /// Starts the WCF service with a singleton instance
        /// </summary>
        /// <param name="ServiceId"></param>
        /// <returns></returns>
        public bool Start(object WebConnectionServer, string ServiceUri)
        {
            // *** if no instance is passed create a COM instance instead
            if (WebConnectionServer == null)
            {
                WebConnectionServer = wwUtils.CreateComInstance("wcDemo.wcDemoServer");
                wwUtils.CallMethodCom(WebConnectionServer, "GetProcessId", 1);
            }

            WebConnectionService Service = new WebConnectionService();
            this.WebConnectionServer = WebConnectionServer;
            Service.WebConnectionServer = WebConnectionServer;             

            try
            {
                Binding Binding = null;
                if (ServiceUri.StartsWith("net.tcp://"))
                    Binding = new NetTcpBinding();
                else if (ServiceUri.StartsWith("net.pipe://"))
                    Binding = new NetNamedPipeBinding();
                else if (ServiceUri.StartsWith("http://"))
                    Binding = new BasicHttpBinding();

                Host = new ServiceHost(Service, new Uri(ServiceUri));
                this.Host.AddServiceEndpoint(typeof(IWebConnectionService), Binding, ServiceUri);                
                Host.Open();
            }
            catch (Exception ex)
            {
                this.ErrorMessage = ex.Message;
                return false;
            }

            return true;
        }

        public bool Stop()
        {
            if (this.WebConnectionServer != null)
            {
                Marshal.FinalReleaseComObject(this.WebConnectionServer);
                this.WebConnectionServer = null;
            }

            if (this.Host != null)
            {
                Host.Close();
                Host = null;
            }
            return true;
        }

        #region IDisposable Members

        public void Dispose()
        {
            this.Stop();
        }

        #endregion
    }
}
